@@ -4,6 +4,7 @@ Module for various derived and user-set constraints
 # Standard library imports
 from collections import namedtuple
 from datetime import datetime
+from functools import partial
 import itertools as it
 import re
 
@@ -11,18 +12,19 @@ import re
 import numpy as np
 import pandas as pd
 
-# Exibit import
+# Exibit imports
+from .sql import query_anon_database
 from .generate.continuous import scale_continuous_column
+from .constants import ORIGINAL_VALUES_DB, ORIGINAL_VALUES_PAIRED, MISSING_DATA_STR
 
 class ConstraintHandler:
     '''
     Keep all internal constraint-handling methods in one place
     with the added bonus of a shared spec_dict object
 
-    Currently, conditional constraints are limited to handling
-    just the na values - "make_nan" and "no_nan". In future, might
-    add more and the code for specific conditional "actions" will
-    go into this class rather than the MissingDataGenerator.
+    Custom constraints are implemented using dedicated functions
+    inside this class, except "make_null" and "make_not_null" which
+    are implemented in the MissingDataGenerator.
     '''
 
     def __init__(self, spec_dict, anon_df):
@@ -39,39 +41,39 @@ class ConstraintHandler:
 
     def process_constraints(self):
         '''
-        Doc string
+        Main function to process constraints. Note that constraints dealing with
+        missing data are handled separately in the missng.py module.
         '''
 
         constraints = self.spec_dict["constraints"]
 
-        boolean_constraints = constraints.get("boolean_constraints", None)
-        conditional_constraints = constraints.get("conditional_constraints", None)
+        basic_constraints = constraints.get("basic_constraints", None)
+        custom_constraints = constraints.get("custom_constraints", None)
 
         # hand over constraints dict / list to the responsible function
         # the function is responsible for implementing the interface. The
         # only restriction on the constraint handling function is that it
         # returns a dataframe.
-        if boolean_constraints:
-            self.output = self.process_boolean_constraints(boolean_constraints)
+        if basic_constraints:
+            self.output = self.process_basic_constraints(basic_constraints)
 
-        if conditional_constraints:
-            self.output = self.process_conditional_constraints(conditional_constraints)
+        if custom_constraints:
+            self.output = self.process_custom_constraints(custom_constraints)
 
         if isinstance(self.output, pd.DataFrame):
             return self.output
         
         return self.input #pragma: no cover
 
-
-    def process_boolean_constraints(self, boolean_constraints):
+    def process_basic_constraints(self, basic_constraints):
         """
-        Adjusts the anonymised dataframe to the boolean constraints specified by
+        Adjusts the anonymised dataframe to the basic boolean constraints specified by
         providing a dependent (to be adjusted) column, an operator, and
         an indepdenent expression (to be adjusted against).
 
         Parameters
         ----------
-        boolean_constraints : list
+        basic_constraints : list
             List of strings with column names escaped with a ~ character
 
         Returns
@@ -84,27 +86,38 @@ class ConstraintHandler:
 
         source = self.input if self.output is None else self.output
 
-        for constraint in boolean_constraints:
+        for constraint in basic_constraints:
             source = self.adjust_dataframe_to_fit_constraint(source, constraint)
 
         return source
 
-    def process_conditional_constraints(self, conditional_constraints):
+    def process_custom_constraints(self, custom_constraints):
         """
-        Adjusts the anonymised dataframe to the boolean constraints specified by
-        providing a dependent (to be adjusted) column, an operator, and
-        an indepdenent expression (to be adjusted against).
+        Handle constraints that are specified in a more flexible way,
+        allowing for targeting specific subsets of data and partitioning
+        the data to enable patterns of seasonality or category-specific 
+        outliers.
+
+        Note that the order of columns specified in the partion section
+        is important and will affect the end result.
+
+        Each custom condition must have a name, but how it's called is not
+        important.
 
         Parameters
         ----------
-        conditional_constraints : list
+        custom_constraints : list
             List with dictionaries that follows the format of:
                 {
-                    Example condition: {
-                        Example column : Example action,
+                    Condition name: {
+                        filter         : A valid pd.eval expression,
+                        partition      : Columns by which to group
+                        targets        : {
+                            Target column : Target action,
+                            }
                     }, 
                 }
-            Same conditions can affect multiple columns, but each column
+            Same action can affect multiple columns, but each column
             can have only one action for ease of writing the YAML spec
 
         Returns
@@ -119,26 +132,38 @@ class ConstraintHandler:
         output_df = source.copy()
 
         dispatch_dict = {
-            "add_outliers" : self.add_outliers
+            "make_outlier"    : self.make_outlier,
+            "sort_ascending"  : partial(self.sort_values, ascending=True),
+            "sort_descending" : partial(self.sort_values, ascending=False),
+            "make_distinct"   : self.make_distinct,
+            "make_same"       : self.make_same,
         }
 
-        for rule, targets in conditional_constraints.items():
-            for target_col, action_str in targets.items():
+        for _, constraint in custom_constraints.items():
+
+            cc_filter = constraint.get("filter", None)
+            cc_partitions = constraint.get("partition", None)
+            cc_targets = constraint.get("targets", dict())
+
+            clean_filter = clean_up_constraint(cc_filter)
+
+            cc_filter_mask = (output_df
+                    .rename(lambda x: x.replace(" ", "__"), axis="columns")
+                    .eval(clean_filter, engine="python"))
+            cc_filter_idx = output_df[cc_filter_mask].index
+
+            # if masked dataframe is empty (no values to adjust), exit loop
+            #remember that instead of adjusting TO the constraint, here the
+            #boolean condition is a FILTER so row matching it must be adjusted
+            if output_df[cc_filter_mask].empty:
+                continue
+
+            for target_col, action_str in cc_targets.items():
                 if (action_func := dispatch_dict.get(action_str, None)):
 
-                    clean_rule = clean_up_constraint(rule)
-                    mask = (output_df
-                        .rename(lambda x: x.replace(" ", "__"), axis="columns")
-                        .eval(clean_rule, engine="python")
-                    )
-
-                    #if masked dataframe is empty (no values to adjust), exit loop
-                    #remember that instead of adjusting TO the constraint, here the
-                    #boolean condition is a FILTER so row matching it must be adjusted
-                    if output_df[mask].empty:
-                        continue
                     # overwrite the original target column with the adjusted one
-                    output_df[target_col] = action_func(mask, output_df[target_col])
+                    output_df.loc[cc_filter_idx, target_col] = action_func(
+                        output_df, cc_filter_idx, target_col, cc_partitions)
 
         
         return output_df
@@ -336,45 +361,99 @@ class ConstraintHandler:
 
     # CONDITIONAL CONSTRAINT FUNCTIONS
     # ================================
-    def add_outliers(self, mask, series, rescale=True):
-        """
-        Create outliers based on the boxplot methodology
+    # Every custom function MUST implement the same call signature
+    # If adding a new function, don't forget to add it to the dispatch dict as well
+    # Each function should also handle missing data and partitioning logic
+    def make_outlier(
+        self, df, filter_idx, target_col, partition_cols=None, rescale=True):
+        '''
+        Make filtered data slice an outlier compared to the rest of the data
+        included in the partition (or entire dataset) using boxplot methodology
 
         Parameters
         ----------
-        mask : boolean index
-            Filter the series to idx that need to be turned into outliers
-        series : pd.Series
-            Original series
-        rescale: boolean
-            Adding outliers can push the series ranges outside the specified bounds
+        df             : pd.DataFrame
+            Unmodified dataframe
+        filter_idx     : pd.Index
+            Index of rows to be modified by the function
+        target_col     : str
+            Column where user wants to add outliers
+        partition_cols : list
+            Columns to group by before computing IQR ranges and outliers
+        rescale        : boolean
+            If true, the whole series including outliers will be rescaled because
+            adding outliers can push the series ranges outside the specified bounds
 
         Returns
         -------
         pd.Series
             New series with outliers substituted for masked values
-        """
+        '''
 
-        q25, q50, q75 = np.percentile(series, [25, 50, 75])
-        iqr = q75 - q25
+        def _within_group_outliers(series):
+            '''
+            Helper function to create outliers within groups - 
+            every value in the new series is an outlier compared
+            to the whole of the group. Watch out for NAs
 
-        if iqr == 0:
+            Outliers are made in one or the other direction based on
+            whether the value is divisible by 2 without remainder.
+            '''
 
-            masked_series = np.where(
-                series[mask] % 2 == 0, series[mask] * 1.3,
-                series[mask] * 0.7
-            )
+            q25, q50, q75 = np.percentile(series, [25, 50, 75])
+            iqr = q75 - q25
+
+            if iqr == 0:
+
+                outlier_series = np.where(
+                    series % 2 == 0, series * 1.3,
+                    series * 0.7
+                )
+
+            else:
+
+                outlier_series = np.where(
+                    series % 2 == 0, (q75 + iqr * 3) + series,
+                    (q25 - iqr * 3) - series
+                )
+
+            return outlier_series
+
+        series = df[target_col].dropna()
+        # make sure that original filtered index reflects the non-null series
+        filter_idx = series.index.intersection(filter_idx)
+
+        if partition_cols is not None:
+
+            partition_cols = [x.strip() for x in partition_cols.split(",") if x]
+            grouped_series = df.dropna(subset=[target_col]).groupby(partition_cols)[target_col]
+            outlier_series = grouped_series.transform(_within_group_outliers)
+            result = series.copy()
+            result.loc[filter_idx] = outlier_series.loc[filter_idx]
 
         else:
 
-            masked_series = np.where(
-                series[mask] >= q50, (q75 + iqr * 3) + series[mask],
-                (q25 - iqr * 3) - series[mask]
-            )
+            q25, q50, q75 = np.percentile(series, [25, 50, 75])
+            iqr = q75 - q25
 
+            if iqr == 0:
 
-        result = series.copy()
-        result.loc[mask] = masked_series
+                masked_series = np.where(
+                    series.loc[filter_idx] % 2 == 0,
+                    series.loc[filter_idx] * 1.3,
+                    series.loc[filter_idx] * 0.7
+                )
+
+            else:
+                # make outliers in the same direction as original values
+                masked_series = np.where(
+                    series.loc[filter_idx] >= q50,
+                    (q75 + iqr * 3) + series.loc[filter_idx],
+                    (q25 - iqr * 3) - series.loc[filter_idx]
+                )
+
+            result = series.copy()
+            result.loc[filter_idx] = masked_series
 
         if rescale:
             col_data = self.spec_dict["columns"][series.name]
@@ -384,9 +463,183 @@ class ConstraintHandler:
 
         return result
 
+    def sort_values(
+        self, df, filter_idx, target_col, partition_cols=None, ascending=True):
+        '''
+        Sort filtered data slice with optional nesting achieved by partitioning
+
+        Parameters
+        ----------
+        df             : pd.DataFrame
+            Unmodified dataframe
+        filter_idx     : pd.Index
+            Index of rows to be modified by the function
+        target_col     : str
+            Column where user wants to add outliers
+        partition_cols : list
+            Columns to group by to achieve nested sort
+        ascending        : boolean
+            Direction of sorting
+
+        Returns
+        -------
+        pd.Series
+            Only the filtered and sorted data slice is returned, not the whole series
+        '''
+
+        if partition_cols is None:
+            
+            new_sorted_series = df.loc[filter_idx, target_col].sort_values(ascending=ascending)
+            new_sorted_series.index = filter_idx
+            return new_sorted_series
+
+        partition_cols = [x.strip() for x in partition_cols.split(",") if x]
+
+        # remember that sorted() defaults to reverse=False so
+        # sorted(False) = ascending; df.sort_values(False) = descending
+        # meaning for sorted() we have to reverse the boolean parameter
+        result = (df
+            .groupby(partition_cols)[target_col]
+            .transform(sorted, reverse=not ascending)
+            .loc[filter_idx]
+        )
+
+        return result
+
+    def make_distinct(self, df, filter_idx, target_col, partition_cols=None):
+        '''
+        Ensure the filtered data slice has distinct values within the specified
+        partition. Paired columns are not supported yet.
+
+        Parameters
+        ----------
+        df             : pd.DataFrame
+            Unmodified dataframe
+        filter_idx     : pd.Index
+            Index of rows to be modified by the function
+        target_col     : str
+            Column where user wants to add outliers
+        partition_cols : list
+            Columns to group by before computing IQR ranges and outliers
+
+        Returns
+        -------
+        pd.Series
+            Only the filtered and sorted data slice is returned, not the whole series
+        '''
+
+        def _make_distinct_within_group(group, original_uniques):
+            '''
+            Helper function with ugly logic. It's possible to assign a distinct value
+            to a duplicate meaning the original value will have to be re-assigned, like
+            in a group A,A,C with uniques A,B,C the first A will stay as A, second A
+            will need a replacement (which comes from the end of the uniques) so gets
+            assigned to C and then C is not in temp_uniques so can't be removed from it.
+            '''
+            # each pass over a group requires a fresh list of uniques to check against
+            # we reverse it because list().pop() works from the end of the list
+            temp_uniques = list(reversed(original_uniques))
+
+            running_set = set()
+            new_group = []
+
+            if not group.duplicated().any():
+                return group
+
+            for x in group:
+                if x not in running_set:
+                    running_set.add(x)
+                    new_group.append(x)
+                    temp_uniques.remove(x)
+                else:
+                    if temp_uniques:
+                        new_unique = temp_uniques.pop()
+                        running_set.add(new_unique)
+                        new_group.append(new_unique)
+                    else:
+                        new_group.append(np.nan)
+
+            return new_group
+
+        orig_vals_in_spec = self.spec_dict["columns"][target_col]["original_values"]
+
+        if isinstance(orig_vals_in_spec, pd.DataFrame):
+            original_uniques = (
+                orig_vals_in_spec[target_col].tolist()
+            )
+
+        elif orig_vals_in_spec == ORIGINAL_VALUES_DB:
+
+            table_name = f'temp_{self.spec_dict["metadata"]["id"]}_{target_col}'
+            original_uniques = query_anon_database(
+                table_name=table_name,
+                column=target_col,
+            )[target_col].tolist()
+
+        elif orig_vals_in_spec == ORIGINAL_VALUES_PAIRED: #pragma: no cover
+            raise ValueError("make_distinct action not supported for paired columns")
+
+        if MISSING_DATA_STR in original_uniques:
+            original_uniques.remove(MISSING_DATA_STR)
+                 
+        if partition_cols is None:
+            
+            filtered_series = df.loc[filter_idx, target_col]
+            result = _make_distinct_within_group(filtered_series, original_uniques)
+            return result
+
+        partition_cols = [x.strip() for x in partition_cols.split(",") if x]
+
+        result = (df
+            .loc[filter_idx]
+            .groupby(partition_cols)[target_col]
+            .transform(_make_distinct_within_group, original_uniques)
+        )
+
+        return result
+
+    def make_same(
+        self, df, filter_idx, target_col, partition_cols=None):
+        '''
+        Force all values in the partition to be the same as the first
+
+        Parameters
+        ----------
+        df             : pd.DataFrame
+            Unmodified dataframe
+        filter_idx     : pd.Index
+            Index of rows to be modified by the function
+        target_col     : str
+            Column where user wants to add outliers
+        partition_cols : list
+            Columns to group by to achieve nested sort
+
+        Returns
+        -------
+        pd.Series
+            Only the filtered data slice with identicl values is returned,
+            not the whole series
+        '''
+
+        if partition_cols is None:
+            
+            repl = df.loc[filter_idx[0], target_col]
+            new_same_series = df.loc[filter_idx, target_col].transform(lambda x: repl)
+            return new_same_series
+
+        partition_cols = [x.strip() for x in partition_cols.split(",") if x]
+
+        result = (df
+            .groupby(partition_cols)[target_col]
+            .transform(lambda x: x.iloc[0])
+            .loc[filter_idx]
+        )
+
+        return result
+
 # EXPORTABLE METHODS
 # ==================
-def find_boolean_columns(df):
+def find_basic_constraint_columns(df):
     '''
     Given a Pandas dataframe, find all numerical column pairs
     that have a relationship that can be described using standard
@@ -457,6 +710,11 @@ def clean_up_constraint(rule_string):
     The solution is to process the constraint first, before passing it to eval,
     not forgetting to rename the dataframe columns with a __ instead of a whitespace
     '''
+
+    # index is always available when doing df.eval or df.query 
+    # and returns a boolean array of True values for each row
+    if rule_string is None:
+        return "index == index"
     
     column_names = re.findall(r"~.*?~", rule_string)
     repl_dict = {"~": "", " ": "__"}
