@@ -15,6 +15,7 @@ import numpy as np
 import pandas as pd
 
 # Exhibit imports
+from .constants import UUID_PLACEHOLDER
 from .specs import newSpec
 from .formatters import parse_original_values
 from .validator import newValidator
@@ -33,6 +34,8 @@ from .generate.continuous import (
                     generate_continuous_column,
                     generate_derived_column)
 
+from .generate.uuids import generate_uuid_column
+
 class newExhibit:
     '''
     The main class encapsulating the demonstrator
@@ -59,10 +62,12 @@ class newExhibit:
     equal_weights  : bool
         Use equal weights and probabilities for all printed column values. Default
         is False.
-    skip_columns   : list
-        List of columns to skip when reading in the data.
-    linked_columns : list
-        List of columns linked with non-hierarchical relationships.
+    skip_columns   : set
+        Set of columns to skip when reading in the data.
+    linked_columns : set
+        Set of columns linked with non-hierarchical relationships.
+    uuid_columns : set
+        Set of columns to treat as having record identifiers.
     verbose        : bool
         Increased verbosity of error messages. Default is False.
 
@@ -82,7 +87,9 @@ class newExhibit:
     def __init__(
         self, command, source, output=None,
         inline_limit=30, equal_weights=False,
-        skip_columns=None, linked_columns=None, verbose=False):
+        skip_columns=None, linked_columns=None, 
+        uuid_columns=None, discrete_columns=None, 
+        verbose=False):
         '''
         Initialise either from the CLI or by instantiating directly
         '''
@@ -96,8 +103,10 @@ class newExhibit:
         self.output = output
         self.inline_limit = inline_limit
         self.equal_weights = equal_weights
-        self.skip_columns = skip_columns
-        self.linked_columns= linked_columns
+        self.skip_columns = skip_columns or set()
+        self.linked_columns= linked_columns or set()
+        self.uuid_columns= uuid_columns or {UUID_PLACEHOLDER}
+        self.discrete_columns = discrete_columns or set()
         self.verbose = verbose
 
         self.spec_dict = None
@@ -120,7 +129,9 @@ class newExhibit:
         else:
             self.df = read_with_date_parser(
                 path=self.source,
-                skip_columns=self.skip_columns)
+                skip_columns=self.skip_columns,
+                discrete_columns=self.discrete_columns,
+                )
 
         return self
 
@@ -130,13 +141,14 @@ class newExhibit:
         only be run after read_data()
         '''
 
-        if not self.df is None:
+        if self.df is not None:
 
             new_spec = newSpec(
                 data=self.df,
                 inline_limit=self.inline_limit,
                 ew=self.equal_weights,
-                user_linked_cols=self.linked_columns
+                user_linked_cols=self.linked_columns,
+                uuid_cols=self.uuid_columns,
                 )
 
             self.spec_dict = new_spec.output_spec_dict()
@@ -233,15 +245,7 @@ class newExhibit:
         # without regard for weights for Missing data which are later adjusted
         # as required.
 
-        #3) CHECK IF DUPLICATES ARE OK
-        if not self.spec_dict["constraints"]["allow_duplicates"]:
-            duplicated_idx = anon_df.duplicated()
-            number_dropped = sum(duplicated_idx)
-            if number_dropped > 0:
-                print(f"WARNING: Deleted {number_dropped} duplicates.")
-                anon_df = anon_df.loc[~duplicated_idx, :].reset_index(drop=True)
-
-        #4) ADD CONTINUOUS VARIABLES TO ANON DF
+        #3) ADD CONTINUOUS VARIABLES TO ANON DF
         # at this point, we don't have any Missing data placeholders (or actual nans)
         # these are added after this step when we can properly account for conditional
         # constraints and other inter-dependencies
@@ -264,20 +268,78 @@ class newExhibit:
                                                     col_name=num_col
             )
 
-        #5) GENERATE MISSING DATA IN ALL COLUMNS
+        #4) GENERATE MISSING DATA IN ALL COLUMNS
         miss_gen = MissingDataGenerator(self.spec_dict, anon_df)
         anon_df = miss_gen.add_missing_data()
 
-        #6) PROCESS BOOLEAN AND CONDITIONAL CONSTRAINTS (IF ANY)
+        #5) GENERATE UUID COLUMNS
+        for uuid_col_name in self.spec_dict["metadata"]["uuid_columns"]:
+
+            dist = self.spec_dict["columns"][uuid_col_name]["frequency_distribution"]
+            
+            # only header included as a placeholder; skip if no frequency information
+            if len(dist) == 1:
+                continue
+
+            uuid_col = generate_uuid_column(
+                uuid_col_name,
+                self.spec_dict["metadata"]["number_of_rows"],
+                self.spec_dict["columns"][uuid_col_name]["miss_probability"],
+                dist,
+                self.spec_dict["metadata"]["random_seed"]
+            )
+
+            anon_df.insert(0, uuid_col_name, uuid_col)
+
+        #6) PROCESS BASIC AND CUSTOM CONSTRAINTS (IF ANY)
         ch = ConstraintHandler(self.spec_dict, anon_df)
         anon_df = ch.process_constraints()
+        # if there are any constraints that affect categorical columns, we need to
+        # re-run the generation of numerical columns because the constraints will
+        # rearrange (sort, make distinct or same) the categorical values leaving
+        # the original numerical values that don't correspond to correct weights
+        constraint_targets = []
+        cat_cols_set = set(self.spec_dict["metadata"]["categorical_columns"])
+
+        if (ccs:=self.spec_dict["constraints"]["custom_constraints"]):
+            
+            for _, cc in ccs.items():
+                constraint_targets.extend(cc["targets"].keys())
+
+        # check if there are common columns between constraint targets and cat_cols
+        if cat_cols_set & set(constraint_targets):
+
+            for num_col in self.spec_dict["metadata"]["numerical_columns"]:
+                
+                # derived columns won't have weight so we ignore them
+                if num_col in self.spec_dict["derived_columns"]: # pragma: no cover
+                    continue
+
+                anon_df[num_col] = generate_continuous_column(
+                                                        spec_dict=self.spec_dict,
+                                                        anon_df=anon_df,
+                                                        col_name=num_col
+                )
 
         #7) GENERATE DERIVED COLUMNS IF ANY ARE SPECIFIED
         for name, calc in self.spec_dict["derived_columns"].items():
             if "Example" not in name:
                 anon_df[name] = generate_derived_column(anon_df, calc)
+
+        #8) CHECK IF DUPLICATES ARE OK
+        # only consider categorical columns (+uuid) for potential duplicates
+        if not self.spec_dict["constraints"]["allow_duplicates"]:
             
-        #8) SAVE THE GENERATED DATASET AS CLASS ATTRIBUTE FOR EXPORT
+            num_cols = self.spec_dict["metadata"]["numerical_columns"]
+            dup_cols = [x for x in anon_df.columns if x not in num_cols]
+
+            duplicated_idx = anon_df.duplicated(subset=dup_cols)
+            number_dropped = sum(duplicated_idx)
+            if number_dropped > 0:
+                print(f"WARNING: Deleted {number_dropped} duplicates.")
+                anon_df = anon_df.loc[~duplicated_idx, :].reset_index(drop=True)
+            
+        #9) SAVE THE GENERATED DATASET AS CLASS ATTRIBUTE FOR EXPORT
         self.anon_df = anon_df
 
     def write_data(self): # pragma: no cover
