@@ -43,15 +43,20 @@ class missingDataTests(unittest.TestCase):
 
     def test_never_null_indices_are_identified(self):
         '''
-        Some cells can't ever have nulls due to conditional constraints
+        Some cells can't ever have nulls due to custom constraints.
+        Filter and Partition fields are optional when defining custom
+        constraints.
         '''
 
         test_dict = {
             "constraints" : {
-                "conditional_constraints": {
-                    "Test == Test" : {    # what to do with non-null rows
-                        "Num" : "no_nan", # ensure a valid value is generated
-                    }
+                "custom_constraints": {
+                    "cc1" : {
+                        "filter"  : "~Test.isna()",
+                        "targets" : {
+                            "Num" : "make_not_null"
+                        }
+                    },
                 }
             }
         }
@@ -63,9 +68,9 @@ class missingDataTests(unittest.TestCase):
 
         test_gen = tm.MissingDataGenerator(test_dict, test_data)
 
-        no_nan_idx = test_gen._find_no_nan_idx()
+        not_null_idx = test_gen._find_not_null_idx()
 
-        result = no_nan_idx[0]
+        result = not_null_idx[0]
 
         assert_series_equal(
             test_data.loc[result],
@@ -104,7 +109,7 @@ class missingDataTests(unittest.TestCase):
                 }
             },
             "constraints" : {
-                "conditional_constraints" : {},
+                "custom_constraints" : {},
                 
                 },
             "linked_columns" : []
@@ -152,7 +157,7 @@ class missingDataTests(unittest.TestCase):
                     }
             },
             "constraints" : {
-                "conditional_constraints" : {},
+                "custom_constraints" : {},
                 },
             "linked_columns" : [
                 (1, ["A", "B"]),
@@ -196,7 +201,7 @@ class missingDataTests(unittest.TestCase):
                     }
             },
             "constraints" : {
-                "conditional_constraints" : {},
+                "custom_constraints" : {},
                 },
             "linked_columns" : [
                 (0, ["A", "C"]),
@@ -212,11 +217,11 @@ class missingDataTests(unittest.TestCase):
 
         self.assertTrue(expected[0], result[0])      
         
-    def test_make_nans_in_columns(self):
+    def test_make_null_constraint_in_columns(self):
         '''
-        When we're adding NANs to categorical columns, the non-null 
+        When we're adding nulls to categorical columns, the non-null 
         numerical values must be re-calulcated and re-scaled because
-        Missing data (NANs identifier in categorical columns) can have
+        Missing data (null identifier in categorical columns) can have
         vastly different weights compared to the old values. However,
         we shouldn't rescaled the whole column anew, just the added values.
         '''
@@ -275,9 +280,12 @@ class missingDataTests(unittest.TestCase):
 
             },
             "constraints" : {
-                "conditional_constraints": {
-                    "A == 'spam'" : {        
-                        "B" : "make_nan",
+                "custom_constraints": {
+                    "cc1" : {
+                        "filter"  : "A == 'spam'",
+                        "targets" : {
+                            "B" : "make_null"
+                        }
                     }
                 }
             },
@@ -303,7 +311,7 @@ class missingDataTests(unittest.TestCase):
 
         assert_frame_equal(result, expected, check_dtype=False)
 
-    def test_no_nans_in_columns(self):
+    def test_not_null_constraint_in_columns(self):
         '''
         Doc string
         '''
@@ -331,9 +339,12 @@ class missingDataTests(unittest.TestCase):
                 }
             },
             "constraints" : {
-                "conditional_constraints": {
-                    "C == C" : {        # what to do with non-null rows
-                        "D" : "no_nan", # ensure a valid value is generated
+                "custom_constraints": {
+                    "cc1" : {
+                        "filter"  : "~C.isna()",
+                        "targets" : {
+                            "D" : "make_not_null"
+                        }
                     }
                 }
             },
@@ -380,7 +391,7 @@ class missingDataTests(unittest.TestCase):
                 },
             },
             "constraints" : {
-                "conditional_constraints" : {},
+                "custom_constraints" : {},
             },
             "linked_columns" : [],
         }
@@ -424,7 +435,7 @@ class missingDataTests(unittest.TestCase):
                 },
             },
             "constraints" : {
-                "conditional_constraints" : {}
+                "custom_constraints" : {}
             },
             "linked_columns" : [],
         }
@@ -504,7 +515,7 @@ class missingDataTests(unittest.TestCase):
                 },
             },
             "constraints" : {
-                "conditional_constraints" : {}
+                "custom_constraints" : {}
             },
             "linked_columns" : [],
             "weights_table" : weights,
