@@ -18,7 +18,7 @@ import numpy as np
 from exhibit.core.utils import package_dir
 from exhibit.db import db_util
 from exhibit.core.constants import MISSING_DATA_STR
-from exhibit.sample.sample import inpatients_anon
+from exhibit.sample.sample import inpatients_anon, uuid_anon
 
 # Module under test
 from exhibit.core import exhibit  as tm
@@ -249,12 +249,12 @@ class referenceTests(unittest.TestCase):
             - removed linked columns from spec
             - removed Scotland from HBs and deleted loc columns
             - changed the totals for stays (100 000) and los (200 000)
-            - changed boolean constraint to los >= stays
+            - changed basic constraint to los >= stays
             - DB is not used at all so no need for ID
 
-        Note that when boolean constraints are added, generated totals can
+        Note that when basic constraints are added, generated totals can
         be different from those set in the spec as target sum is enforced
-        BEFORE boolean constraints are adjusted.
+        BEFORE basic constraints are adjusted.
         '''
 
 
@@ -484,7 +484,7 @@ class referenceTests(unittest.TestCase):
                     {"anonymising_set":"mountains.peak"}
                 },
             "constraints": {
-                "boolean_constraints" : {}
+                "basic_constraints" : {}
             }
         }
                  
@@ -686,6 +686,40 @@ class referenceTests(unittest.TestCase):
 
         self.assertCountEqual(temp_df["loc_name"].unique(), list("ABCDE"))
 
+    def test_reference_uuid_data(self):
+        '''
+        What this reference test is covering:
+            - uuid column type
+            - make_distinct, make_same and sorting custom actions
+            - no db
+        '''
+        
+        args = dict(
+                command="fromspec",
+                source=Path(package_dir("sample", "_spec", "uuid_demo.yml")),
+                verbose=True,
+                skip_columns=[]
+            )
+
+        xA = tm.newExhibit(**args)
+        xA.read_spec()
+        if xA.validate_spec():
+            xA.execute_spec()
+
+        #sort column names to make sure they are the same
+        uuid_anon.sort_index(axis=1, inplace=True)
+        xA.anon_df.sort_index(axis=1, inplace=True)
+
+        # there is a quirk of how int is cast on Windows and Unix: int32 vs int64
+        # see SO answer:
+        # Why do Pandas integer `dtypes` not behave the same on Unix and Windows?
+        assert_frame_equal(
+            left=uuid_anon,
+            right=xA.anon_df,
+            check_exact=False,
+            check_dtype=False
+        )
+
 if __name__ == "__main__" and __package__ is None:
     #overwrite __package__ builtin as per PEP 366
     __package__ = "exhibit"
